#define die_dberr(t,f...) { LFATAL(f); exit(EXIT_DBERR); }
#define die_dataerr(f...) { LFATAL(f); exit(EXIT_DATAERR); }
#define die_syserr() { LFATAL("Fatal system error : %s",strerror(errno)); exit(EXIT_SYSTEM); }
#define TIGERLEN 24
#define MAX_POSIX_FILENAME_LEN 256
#define MAX_FUSE_BLKSIZE 131072

typedef struct {
    unsigned long long inode;
    unsigned long long blocknr;
} INOBNO;

typedef struct {
    //unsigned int snapshotnr;
    unsigned long long dirnode;
    unsigned long long inode;
} DINOINO;

typedef struct {
    unsigned long size;
    unsigned char *data;
} DBT;

typedef struct {
    struct stat stbuf;
    char filename[MAX_POSIX_FILENAME_LEN];
} DDSTAT;

typedef struct {
    unsigned long long blocknr;
    unsigned char blockdata[MAX_FUSE_BLKSIZE];
} BLKCACHE;

typedef struct {
    struct stat stbuf;
    unsigned int updated;
    unsigned long long blocknr;
    unsigned int opened;
    unsigned long long deduplicated;
    unsigned long long lzo_compressed_size;
    char filename[MAX_POSIX_FILENAME_LEN];
} MEMDDSTAT;

typedef struct {
    compr *compressed;
    const char *blockdata;
    unsigned long long blocknr;
    unsigned int offsetblock;
    size_t bsize;
    unsigned long long inode;
    bool sparse;
    unsigned char *stiger;
    unsigned char *blockfiller;
} BLKDTA;

DBT *check_block_exists(INOBNO);
int commitBlock(unsigned long long, const char *, const char *,
                unsigned long long);
unsigned long long readBlock(unsigned long long, const char *, char *,
                             unsigned long long);
int dbstat(const char *, struct stat *);
void formatfs();
unsigned long long get_next_inode();
void bin_write_dbdata(TCHDB *, void *, int, void *, int);
void asc_write_dbdata(TCHDB *, unsigned char *, unsigned char *);
void mbin_write_dbdata(TCMDB *, void *, int, void *, int);
void dbmknod(const char *, mode_t, char *, dev_t);
void getLock();
void releaseLock();
void DBTfree(DBT *);
void delete_key(TCHDB *, void *, int);
unsigned long long getInUse(unsigned char *);
void tc_open(bool);
void tc_close(bool);
DBT *search_dbdata(TCHDB *, void *key, int);
char *hash(char *, int);
void updateInUse(unsigned char *, int, unsigned long long);
void hash_update_filesize(MEMDDSTAT *, unsigned long long);
void update_filesize(unsigned long long, unsigned long long, unsigned int,
                    unsigned long long, bool, unsigned int, unsigned int);
void addBlock(BLKDTA *);
void update_block(const char *, unsigned long long,
                  unsigned int, unsigned long long, unsigned long long,
                  unsigned char *);
char *s_dirname(char *);
char *s_basename(char *);
void write_file_ent(const char *, unsigned long long, mode_t mode, char *,
                  dev_t);
int unlink_file(const char *);
int fs_mkdir(const char *, mode_t);
int fs_truncate(struct stat *, off_t, char *);
int fs_rmdir(const char *);
unsigned long long get_inode(const char *);
int fs_readdir(const char *, void *, fuse_fill_dir_t, off_t,
               struct fuse_file_info *);
void btasc_curwrite_dbdata(TCBDB *, BDBCUR *, unsigned char *);
void btasc_write_dbdata(TCBDB *, char *, char *);
int fs_link(char *, char *);
int fs_symlink(char *, char *);
int fs_readlink(const char *, char *, size_t);
int fs_rename(const char *, const char *, struct stat);
int fs_rename_link(const char *, const char *, struct stat);
char *fs_search_topdir(char *);
void btbin_write_dbdata(TCBDB *, void *, int, void *, int);
void btbin_curwrite_dbdata(TCBDB *, BDBCUR *, char *, int);
void ddstatfree(DDSTAT *);
DBT *create_ddbuf(struct stat, char *);
DDSTAT *value_to_ddstat(char *, int);
unsigned long long has_nodes(unsigned long long);
void fil_fuse_info(DDSTAT *, void *, fuse_fill_dir_t,
                   struct fuse_file_info *);
void bt_curwrite(TCBDB *, char *, char *);
int bt_entry_exists(TCBDB *, void *, int, void *, int);
DDSTAT *dnode_bname_to_inode(void *, int, char *);
int count_dirlinks(void *, int);
MEMDDSTAT *value_tomem_ddstat(char *, int);
DBT *search_memhash(TCMDB *, void *, int);
DBT *create_mem_ddbuf(MEMDDSTAT *);
void update_filesize_onclose(unsigned long long);
void mdelete_key(TCMDB *, void *, int);
void update_cache(unsigned long long, struct stat *);
int update_stat(char *path, struct stat *);
void memddstatfree(MEMDDSTAT *);
int update_parent_time(char *);
void tc_defrag();
unsigned char *binhash(unsigned char *, int);
void dta_mutex_init(int);
void release_dta_lock(int);
void release_write_lock();
void dta_lock(int);
void thread_lock();
void thread_lock_release();
void release_all_dta_lock();
void all_dta_lock();
void write_lock();
void tiger_lock();
void open_lock();
void release_tiger_lock();
void release_open_lock();
void get_qempty_lock();
void get_qdta_lock();
void get_moddb_lock();
void release_moddb_lock();
void release_qempty_lock();
void wait_io_pending(unsigned long long);
void write_nfi(unsigned long long);
void btbin_write_dup(TCBDB *, void *, int, void *, int);
void *btsearch_keyval(TCBDB *, void *, int, void *, int);
int inode_block_pending(unsigned long long, unsigned long long);
DBT *try_block_cache(unsigned long long, unsigned long long, unsigned int);
void flush_dta_queue();
void add_blk_to_cache(unsigned long long, unsigned long long, unsigned char *);
void qdta(unsigned char *, unsigned char *, unsigned int);
void comprfree(compr *);
void loghash(char *, unsigned char *);
